header {* non linear polynomials *}

theory NLPolynomial
imports Matrix (* only used for order_pair, etc., no matrices are used *)
begin

subsection {*
polynomials represented as trees
*}
datatype ('v,'a)tpoly = PVar 'v | PNum 'a | PSum "('v,'a)tpoly list" | PMult "('v,'a)tpoly list"

types ('v,'a)assign = "'v \<Rightarrow> 'a"

fun eval_tpoly :: "('v,'a :: semiring_1)assign \<Rightarrow> ('v,'a)tpoly \<Rightarrow> 'a"
where "eval_tpoly \<alpha> (PVar x) = \<alpha> x"
   |  "eval_tpoly \<alpha> (PNum a) = a"
   |  "eval_tpoly \<alpha> (PSum []) = 0"
   |  "eval_tpoly \<alpha> (PSum (p # ps)) = eval_tpoly \<alpha> p + eval_tpoly \<alpha> (PSum ps)"
   |  "eval_tpoly \<alpha> (PMult []) = 1"
   |  "eval_tpoly \<alpha> (PMult (p # ps)) = eval_tpoly \<alpha> p * eval_tpoly \<alpha> (PMult ps)"

subsection {*  polynomials represented in normal form *}
text {*
  The internal representation of polynomials is: sum of products of monomials with coefficients.
  all coefficients are non-zero, all monomials are different
*}
 

text {* Definition of type @{text monom} *}


types 'v monom = "('v \<times> nat)list" 
(* [(x,n),(y,m)] represent x^n * y^m 
   invariants: all powers are \<ge> 1
               each variable occurs at most once
   hence: [(x,1),(y,2),(x,2)] will not occur, but [(x,3),(y,2)]
          [(x,1),(y,0)] will not occur, but [(x,1)]
*)


fun eval_monom :: "('v,'a :: comm_semiring_1)assign \<Rightarrow> ('v monom) \<Rightarrow> 'a"
where "eval_monom \<alpha> [] = 1"
   |  "eval_monom \<alpha> ((x,p) # m) = eval_monom \<alpha> m * (\<alpha> x)^p"

lemma eval_monom_list[simp]: "eval_monom \<alpha> (m @ n) = eval_monom \<alpha> m * eval_monom \<alpha> n"
  by (induct m, auto simp: field_simps)

text {*
equality of monomials should be able to identify x^2*y with y*x^2,
essentially, it checks for permutations. Checking of permutations 
suffices, since in x^2*x^1 = x^3, the left-hand side should not
be constructed due to the invariant, that every variable occurs at
most once in a monomial.
*}
fun eq_monom :: "'v monom \<Rightarrow> 'v monom \<Rightarrow> bool" (infix "=m" 51)
where "[] =m n = (n = [])" 
  |   "(x,p) # m =m n = (case extract (\<lambda> yq. fst yq = x) n of
                           None \<Rightarrow> False
                         | Some (n1,(_,q),n2) \<Rightarrow> p = q \<and> m =m (n1 @ n2))"

lemma eq_monom: "m =m n \<Longrightarrow> eval_monom \<alpha> m = eval_monom \<alpha> n"
proof (induct m arbitrary: n, simp)
  case (Cons xp m) note mCons = this
  show ?case
  proof (cases xp)
    case (Pair x p)
    show ?thesis 
    proof (cases "extract (\<lambda> yq. fst yq = x) n")
      case None
      with Cons Pair show ?thesis by auto
    next
      case (Some res)
      obtain n1 yq n2 where "res = (n1,yq,n2)" by (cases res, auto)
      then obtain y q where res: "res = (n1,(y,q),n2)" by (cases yq, auto)
      from extract_Some[OF Some[simplified res]] mCons(2)  Some Pair res have n: "n = n1 @ (x,p) # n2" and rec: "m =m (n1 @ n2)" by auto
      show ?thesis by (simp add: Pair mCons(1)[OF rec] n field_simps)
    qed
  qed
qed

declare eq_monom.simps[simp del]

fun monom_mult :: "'v monom \<Rightarrow> 'v monom \<Rightarrow> 'v monom"
where "monom_mult [] n = n"
  |   "monom_mult ((x,p) # m) n = (case extract (\<lambda> yq. fst yq = x) n of
                           None \<Rightarrow> (x,p) # monom_mult m n 
                         | Some (n1,(_,q),n2) \<Rightarrow> (x,p+q) # monom_mult m (n1 @ n2))"


lemma monom_mult[simp]: "eval_monom \<alpha> (monom_mult m n) = eval_monom \<alpha> m * eval_monom \<alpha> n"
proof (induct m arbitrary: n, simp)
  case (Cons xp m) note mCons = this
  show ?case
  proof (cases xp)
    case (Pair x p)
    show ?thesis 
    proof (cases "extract (\<lambda> yq. fst yq = x) n")
      case None
      with Cons Pair show ?thesis by (auto simp: field_simps)
    next
      case (Some res)
      obtain n1 yq n2 where "res = (n1,yq,n2)" by (cases res, auto)
      then obtain y q where res: "res = (n1,(y,q),n2)" by (cases yq, auto)
      from extract_Some[OF Some[simplified res]] Some Pair res have n: "n = n1 @ (x,q) # n2" 
        and rec: "monom_mult (xp # m) n = (x, p+q) # monom_mult m (n1 @ n2)" by auto
      show ?thesis by (simp only: rec, simp add: Pair mCons[of "n1 @ n2"] n field_simps power_add)
    qed
  qed
qed

declare monom_mult.simps[simp del]



types ('v,'a)poly = "('v monom \<times> 'a)list"
(* a list of monomials with their coefficient;
   invariants: all coefficients are non-zero;
               no two =m equivalent monomials occur in the list
*)

abbreviation eval_monomc where "eval_monomc \<alpha> mc \<equiv> eval_monom \<alpha> (fst mc) * (snd mc)"

fun eval_poly :: "('v,'a :: comm_semiring_1)assign \<Rightarrow> ('v,'a)poly \<Rightarrow> 'a"
where "eval_poly \<alpha> [] = 0"
   |  "eval_poly \<alpha> (mc # p) = eval_monomc \<alpha> mc + eval_poly \<alpha> p"


fun poly_add :: "('v,'a)poly \<Rightarrow> ('v,'a :: semiring_0)poly \<Rightarrow> ('v,'a)poly"
where "poly_add [] q = q"
   |  "poly_add ((m,c) # p) q = (case extract (\<lambda> mc. fst mc =m m) q of
                                None \<Rightarrow> (m,c) # poly_add p q
                              | Some (q1,(_,d),q2) \<Rightarrow> if (c+d = 0) then poly_add p (q1 @ q2) else (m,c+d) # poly_add p (q1 @ q2))"

lemma eval_poly_append[simp]: "eval_poly \<alpha> (mc1 @ mc2) = eval_poly \<alpha> mc1 + eval_poly \<alpha> mc2"
  by (induct mc1, auto simp: field_simps)

lemma poly_add[simp]: "eval_poly \<alpha> (poly_add p q) = eval_poly \<alpha> p + eval_poly \<alpha> q"
proof (induct p arbitrary: q, simp)
  case (Cons mc p)
  obtain m c where mc: "mc = (m,c)" by (cases mc, auto)
  show ?case
  proof (cases "extract (\<lambda> mc. fst mc =m m) q")
    case None
    show ?thesis by (simp add: Cons[of q] mc None field_simps)
  next
    case (Some res)
    obtain q1 md q2 where res: "res = (q1,md,q2)" by (cases res, auto)
    from extract_Some[OF Some[simplified res]] res obtain m' d where q: "q = q1 @ (m',d) # q2" and m': "m' =m m" and res: "res = (q1,(m',d),q2)" by (cases md, auto)
    {
      fix x
      assume c: "c + d = 0"
      have "c * x + d * x = (c + d) * x" by (auto simp: field_simps)
      also have "\<dots> = 0 * x" by (simp only: c)
      finally have "c * x + d * x = 0" by simp
    } note id = this
    show ?thesis 
      by (simp add: Cons[of "q1 @ q2"] mc Some res, simp only: q, simp add: eq_monom[OF m'] field_simps, auto simp: field_simps id)
  qed
qed

declare poly_add.simps[simp del]

(* TODO: check whether c * d is definitely non-zero, otherwise invariant is destroyed.
   (note that c and d are non-zero) *)
fun monom_mult_poly :: "('v monom \<times> 'a) \<Rightarrow> ('v,'a :: semiring_0)poly \<Rightarrow> ('v,'a)poly"
where "monom_mult_poly (m,c) p = (if c = 0 then [] else map (\<lambda> (m',d). (monom_mult m m', c * d)) p)"

lemma monom_mult_poly[simp]: "eval_poly \<alpha> (monom_mult_poly mc p) = eval_monomc \<alpha> mc * eval_poly \<alpha> p"
proof (cases mc)
  case (Pair m c)
  show ?thesis
  proof (cases "c = 0")
    case True
    with Pair show ?thesis by simp
  next
    case False
    have "eval_poly \<alpha> (monom_mult_poly mc p) = eval_poly \<alpha> (map (\<lambda> (m',d). (monom_mult m m', c * d)) p)"
      by (simp add: Pair False)
    also have "\<dots> = eval_monom \<alpha> m * c * eval_poly \<alpha> p"
    proof (induct p, simp)
      case (Cons nd q)
      obtain n d where nd: "nd = (n,d)" by (cases nd, auto)
      show ?case
        by (simp add: nd Cons, simp only: field_simps)
    qed
    finally show ?thesis using Pair by simp
  qed
qed

declare monom_mult_poly.simps[simp del]

fun poly_mult :: "('v,'a :: semiring_0)poly \<Rightarrow> ('v,'a)poly \<Rightarrow> ('v,'a)poly" 
where "poly_mult [] q = []"
   |  "poly_mult (mc # p) q = poly_add (monom_mult_poly mc q) (poly_mult p q)"

lemma poly_mult[simp]: "eval_poly \<alpha> (poly_mult p q) = eval_poly \<alpha> p * eval_poly \<alpha> q"
proof (induct p, simp)
  case (Cons mc p)
  thus ?case
    by (simp add: field_simps)
qed

declare poly_mult.simps[simp del]

definition zero_poly :: "('v,'a)poly"
where "zero_poly \<equiv> []"

definition one_poly :: "('v,'a :: semiring_1)poly"
where "one_poly \<equiv> [([],1)]"

lemma poly_zero_add: "poly_add zero_poly p = p" unfolding zero_poly_def using poly_add.simps by auto

lemma poly_zero_mult: "poly_mult zero_poly p = zero_poly" unfolding zero_poly_def using poly_mult.simps by auto

lemma poly_one_mult: "poly_mult one_poly p = p" 
proof - 
  have "poly_add p [] = p" 
    by (induct p, auto simp: poly_add.simps)
  thus ?thesis unfolding one_poly_def
    by (simp add: poly_mult.simps poly_add.simps monom_mult_poly.simps monom_mult.simps)
qed

text {* let us state that we compute correct by linking to tpolys *}
fun
  poly_of :: "('v,'a :: comm_semiring_1)tpoly \<Rightarrow> ('v,'a)poly"
where "poly_of (PNum i) = (if i = 0 then [] else [([],i)])"
    | "poly_of (PVar x) = [([(x,1)],1)]"
    | "poly_of (PSum []) = zero_poly" 
    | "poly_of (PSum (p # ps)) = (poly_add (poly_of p) (poly_of (PSum ps)))"
    | "poly_of (PMult []) = one_poly" 
    | "poly_of (PMult (p # ps)) = (poly_mult (poly_of p) (poly_of (PMult ps)))"

lemma poly_of: "eval_poly \<alpha> (poly_of p) = eval_tpoly \<alpha> p"
by (induct p, (simp add: zero_poly_def one_poly_def)+)

text {* and continue to work on polynomials in normal form *}
definition eq_poly :: "('v,'a :: comm_semiring_1)poly \<Rightarrow> ('v,'a)poly \<Rightarrow> bool" (infix "=p" 51)
where "p =p q \<equiv> \<forall> \<alpha>. eval_poly \<alpha> p = eval_poly \<alpha> q"

lemma eq_poly_refl[simp]: "p =p p" unfolding eq_poly_def by auto

lemma eq_poly_trans[trans]: "\<lbrakk>p1 =p p2; p2 =p p3\<rbrakk> \<Longrightarrow> p1 =p p3"
using assms unfolding eq_poly_def by auto

lemma poly_add_comm: "poly_add p q =p poly_add q p" unfolding eq_poly_def by (auto simp: field_simps)

lemma poly_add_assoc: "poly_add p1 (poly_add p2 p3) =p poly_add (poly_add p1 p2) p3" unfolding eq_poly_def by (auto simp: field_simps)

lemma poly_mult_comm: "poly_mult p q =p poly_mult q p" unfolding eq_poly_def by (auto simp: field_simps)

lemma poly_mult_assoc: "poly_mult p1 (poly_mult p2 p3) =p poly_mult (poly_mult p1 p2) p3" unfolding eq_poly_def by (auto simp: field_simps)

lemma poly_distrib: "poly_mult p (poly_add q1 q2) =p poly_add (poly_mult p q1) (poly_mult p q2)" unfolding eq_poly_def by (auto simp: field_simps)

text {* powers of polynomials *}
fun poly_power :: "('v,'a :: comm_semiring_1)poly \<Rightarrow> nat \<Rightarrow> ('v,'a)poly"
where "poly_power _ 0 = [([],1)]"
  | "poly_power p (Suc n) = poly_mult p (poly_power p n)"

lemma poly_power[simp]: "eval_poly \<alpha> (poly_power p n) = (eval_poly \<alpha> p) ^ n"
  by (induct n, auto)

declare poly_power.simps[simp del]

text {* substitutions in polynomials *}
fun monom_subst :: "('v \<Rightarrow> ('w,'a :: comm_semiring_1)poly) \<Rightarrow> 'v monom \<Rightarrow> ('w,'a)poly"
where "monom_subst \<sigma> [] = one_poly"
  | "monom_subst \<sigma> ((x,p) # m) = poly_mult (poly_power (\<sigma> x) p) (monom_subst \<sigma> m)"

lemma monom_subst[simp]: "eval_poly \<alpha> (monom_subst \<sigma> m) = eval_monom (\<lambda> v. eval_poly \<alpha> (\<sigma> v)) m"
  by (induct m, simp add: one_poly_def, auto simp: field_simps)

declare monom_subst.simps[simp del]

fun poly_subst :: "('v \<Rightarrow> ('w,'a :: comm_semiring_1)poly) \<Rightarrow> ('v,'a)poly \<Rightarrow> ('w,'a)poly"
where "poly_subst \<sigma> [] = zero_poly"
  | "poly_subst \<sigma> ((m,c) # p) = poly_add (poly_mult [([],c)] (monom_subst \<sigma> m)) (poly_subst \<sigma> p)"

lemma poly_subst: "eval_poly \<alpha> (poly_subst \<sigma> p) = eval_poly (\<lambda> v. eval_poly \<alpha> (\<sigma> v)) p"
  by (induct p, simp add: zero_poly_def, auto simp: field_simps)

(* check whether a variable occurs in p *)
definition poly_var :: "('v,'a)poly \<Rightarrow> 'v \<Rightarrow> bool"
where "poly_var p v = v mem (concat_map (map fst o fst) p)"

lemma poly_var: assumes pv: "\<not> poly_var p v" and diff: "\<And> w. v \<noteq> w \<Longrightarrow> f w = g w"
  shows "poly_subst f p = poly_subst g p" 
using pv 
proof (induct p, simp)
  case (Cons mc p)
  from Cons(2) have a: "\<not> poly_var p v" and b: "v \<notin> set ((map fst o fst) mc)" unfolding poly_var_def by (auto simp: mem_iff)
  from Cons(1)[OF a] have rec: "poly_subst f p = poly_subst g p" .
  show ?case
  proof (cases mc)
    case (Pair m c)
    with b have nmem: "v \<notin> set (map fst m)" by auto
    hence "monom_subst f m = monom_subst g m"
    proof (induct m, simp add: monom_subst.simps)
      case (Cons wn m)
      hence rec: "monom_subst f m = monom_subst g m" and neq: "v \<noteq> fst wn" by auto
      show ?case
      proof (cases wn)
        case (Pair w n)
        with diff[OF neq] rec show ?thesis by (auto simp: monom_subst.simps)
      qed
    qed
    with rec Pair show ?thesis by auto
  qed
qed
           
declare poly_subst.simps[simp del]



subsection "Polynomial Orders"

class poly_carrier = ordered_semiring_1 + comm_semiring_1

definition pos_assign :: "('v,'a :: ordered_semiring_0)assign \<Rightarrow> bool"
where "pos_assign \<alpha> = (\<forall> x. \<alpha> x \<ge> 0)"

definition poly_ge :: "('v,'a :: poly_carrier)poly \<Rightarrow> ('v,'a)poly \<Rightarrow> bool" (infix "\<ge>p" 51)
where "p \<ge>p q = (\<forall> \<alpha>. pos_assign \<alpha> \<longrightarrow> ge (eval_poly \<alpha> p) (eval_poly \<alpha> q))"

lemma poly_ge_refl[simp]: "p \<ge>p p"
unfolding poly_ge_def using ge_refl by auto

lemma poly_ge_trans[trans]: "\<lbrakk>p1 \<ge>p p2; p2 \<ge>p p3\<rbrakk> \<Longrightarrow> p1 \<ge>p p3"
using assms unfolding poly_ge_def using ge_trans by blast


lemma pos_assign_monom: fixes \<alpha> :: "('v,'a :: poly_carrier)assign"
  assumes pos: "pos_assign \<alpha>"
  shows "ge (eval_monom \<alpha> m) 0"
proof (induct m, simp add: one_ge_zero)
  case (Cons xp m)
  show ?case
  proof (cases xp)
    case (Pair x p)
    from pos[unfolded pos_assign_def] have ge: "ge (\<alpha> x) 0" by simp
    have ge: "ge (\<alpha> x ^ p) 0"
    proof (induct p, simp add: one_ge_zero)
      case (Suc p)
      from ge_trans[OF times_left_mono[OF ge Suc] times_right_mono[OF ge_refl ge]]
      show ?case by (simp add: field_simps)
    qed
    from ge_trans[OF times_right_mono[OF Cons ge] times_left_mono[OF ge_refl Cons]]
    show ?thesis
      by (simp add: Pair)
  qed
qed

lemma pos_assign_poly:   assumes pos: "pos_assign \<alpha>"
  and p: "p \<ge>p zero_poly"
  shows "ge (eval_poly \<alpha> p) 0"
proof -
  from p[unfolded poly_ge_def zero_poly_def] pos 
  show ?thesis by auto
qed


lemma poly_add_ge_mono: assumes "p1 \<ge>p p2" shows "poly_add p1 q \<ge>p poly_add p2 q"
using assms unfolding poly_ge_def by (auto simp: field_simps plus_left_mono)

lemma poly_mult_ge_mono: assumes "p1 \<ge>p p2" and "q \<ge>p zero_poly"
  shows "poly_mult p1 q \<ge>p poly_mult p2 q"
using assms unfolding poly_ge_def zero_poly_def by (auto simp: times_left_mono)

definition poly_weak_mono :: "'w \<Rightarrow> ('v,'a :: poly_carrier)poly \<Rightarrow> bool"
where "poly_weak_mono type p \<equiv> \<forall> (f :: 'v \<Rightarrow> ('w,'a)poly) g. (\<forall> x. f x \<ge>p g x \<and> g x \<ge>p zero_poly) \<longrightarrow> poly_subst f p \<ge>p poly_subst g p"

fun check_poly_ge :: "('v,'a :: ordered_semiring_0)poly \<Rightarrow> ('v,'a)poly \<Rightarrow> bool"
where "check_poly_ge [] q = list_all (\<lambda> (_,d). ge 0 d) q"
    | "check_poly_ge ((m,c) # p) q = (case extract (\<lambda> nd. fst nd =m m) q of
                                         None \<Rightarrow> ge c 0 \<and> check_poly_ge p q
                                       | Some (q1,(_,d),q2) \<Rightarrow> ge c d \<and> check_poly_ge p (q1 @ q2))"

lemma check_poly_ge: fixes p :: "('v,'a :: poly_carrier)poly"
  shows "check_poly_ge p q \<Longrightarrow> p \<ge>p q"
proof (induct p arbitrary: q)
  case Nil
  hence "\<forall> (n,d) \<in> set q. ge 0 d" using list_all_iff[of _ q] by auto
  hence "[] \<ge>p q" 
  proof (induct q, simp add: poly_ge_refl)
    case (Cons nd q)
    hence rec: "[] \<ge>p q" by simp
    show ?case
    proof (cases nd)
      case (Pair n d)
      with Cons have ge: "ge 0 d" by auto
      show ?thesis 
      proof (simp add: Pair, unfold poly_ge_def, intro allI impI, simp)
        fix \<alpha> :: "('v,'a)assign"
        assume pos: "pos_assign \<alpha>"
        have ge: "ge 0 (eval_monom \<alpha> n * d)"
          using times_right_mono[OF pos_assign_monom[OF pos, of n] ge] by simp
        from rec[unfolded poly_ge_def] pos have ge2: "ge 0 (eval_poly \<alpha> q)" by auto
        show "ge 0 (eval_monom \<alpha> n * d + eval_poly \<alpha> q)" using ge_trans[OF plus_left_mono[OF ge] plus_right_mono[OF ge2]]
          by simp
      qed
    qed
  qed
  thus ?case by simp
next
  case (Cons mc p)
  obtain m c where mc: "mc = (m,c)" by (cases mc, auto)
  show ?case
  proof (cases "extract (\<lambda> mc. fst mc =m m) q")
    case None
    with Cons(2) have rec: "check_poly_ge p q" and c: "ge c 0" using mc by auto
    from Cons(1)[OF rec] have rec: "p \<ge>p q" .
    show ?thesis 
    proof (simp add: mc, unfold poly_ge_def, intro allI impI, simp)
      fix \<alpha> :: "('v,'a)assign"
      assume pos: "pos_assign \<alpha>"
      have ge: "ge (eval_monom \<alpha> m * c) 0"
        using times_right_mono[OF pos_assign_monom[OF pos, of m] c] by simp
      from rec have pq: "eval_poly \<alpha> p \<ge> eval_poly \<alpha> q" unfolding poly_ge_def using pos by auto
      show "eval_monom \<alpha> m * c + eval_poly \<alpha> p \<ge> eval_poly \<alpha> q"
        using ge_trans[OF plus_left_mono[OF ge] plus_right_mono[OF pq]] by simp
    qed
  next
    case (Some res)
    obtain q1 md q2 where "res = (q1,md,q2)" by (cases res, auto)
    with extract_Some[OF Some[simplified this]] obtain m' d where q: "q = q1 @ (m',d) # q2" and m': "m' =m m" and res: "res = (q1,(m',d),q2)" 
      by (cases md, auto)
    from Cons(2) Some mc res have rec: "check_poly_ge p (q1 @ q2)" and c: "ge c d" by auto
    from Cons(1)[OF rec] have p: "p \<ge>p q1 @ q2" .
    show ?thesis
    proof (simp add: mc, unfold poly_ge_def, intro allI impI)
      fix \<alpha> :: "('v,'a)assign"
      assume pos: "pos_assign \<alpha>"
      have ge: "ge (eval_monom \<alpha> m * c) (eval_monom \<alpha> m' * d)"
        using times_right_mono[OF pos_assign_monom[OF pos, of m] c] 
          eq_monom[OF m', of \<alpha>] by simp
      from p have ge2: "ge (eval_poly \<alpha> p) (eval_poly \<alpha> (q1 @ q2))" unfolding poly_ge_def using pos by auto
      show "ge (eval_poly \<alpha> ((m,c) # p)) (eval_poly \<alpha> q)" using ge_trans[OF plus_left_mono[OF ge] plus_right_mono[OF ge2]]
        by (simp add: q field_simps)
    qed
  qed
qed

declare check_poly_ge.simps[simp del]

definition check_poly_weak_mono :: "('v,'a :: ordered_semiring_0)poly \<Rightarrow> bool"
where "check_poly_weak_mono p \<equiv> list_all (\<lambda> (m,c). ge c 0) p"

lemma monom_subst_mono: fixes \<alpha> :: "('v,'a :: poly_carrier)assign"
  assumes pos: "pos_assign \<alpha>"
  and fg: "\<And> v. f v \<ge>p g v"
  and g: "\<And> v. g v \<ge>p zero_poly"
  shows "ge (eval_poly \<alpha> (monom_subst f m)) (eval_poly \<alpha> (monom_subst g m))" 
proof -
  let ?f = "\<lambda> v. eval_poly \<alpha> (f v)"
  let ?g = "\<lambda> v. eval_poly \<alpha> (g v)"
  have fg: "\<And> v. ?f v \<ge> ?g v" and g: "\<And> v. ?g v \<ge> 0"  using fg g pos unfolding poly_ge_def zero_poly_def by auto
  have "(eval_monom ?f m \<ge> eval_monom ?g m) \<and> (eval_monom ?g m \<ge> 0)" 
  proof (induct m, simp add: ge_refl one_ge_zero)
    case (Cons vn m)
    show ?case 
    proof (cases vn)
      case (Pair v n)
      from Cons have one: "eval_monom ?f m \<ge> eval_monom ?g m" and one': "eval_monom ?g m \<ge> 0" by simp+
      { (* perhaps extract this proof *)
        fix a b :: 'a and n :: nat
        assume ge: "ge a b" and ze: "ge b 0"
        have "ge (a^n) (b^n) \<and> ge (b^n) 0"
        proof (induct n, simp add: ge_refl one_ge_zero)
          case (Suc n)
          hence rec1: "a ^ n \<ge> b ^ n" and rec2: "b ^ n \<ge> 0" by auto
          from times_left_mono[OF ze rec2] have one: "b ^ Suc n \<ge> 0" by (auto simp: field_simps)
          have "a * a ^n \<ge> a * b ^ n" using times_left_mono[OF ge_trans[OF ge ze] rec1] by (simp add: field_simps)
          also have "\<dots> \<ge> b * b ^ n" using times_left_mono[OF rec2 ge] .
          finally  show ?case by (simp only: one, simp)
        qed
      } note power_mono = this
      from power_mono[OF fg g]
      have two: "?f v^n \<ge> ?g v^n" and two': "?g v^n \<ge> 0" by auto
      from times_left_mono[OF ge_trans[OF one one'] two] have "eval_monom ?f m * ?f v^n \<ge> eval_monom ?f m * ?g v^n" by (auto simp: field_simps)
      also have "\<dots> \<ge> eval_monom ?g m * ?g v^n" using times_left_mono[OF two' one] .
      finally have res1: "eval_monom ?f m * ?f v^n \<ge> eval_monom ?g m * ?g v^n" .
      from times_left_mono[OF two' one'] have res2: "eval_monom ?g m * ?g v^n \<ge> 0" by simp
      show ?thesis by (simp add: Pair res1 res2)
    qed
  qed
  thus ?thesis by simp
qed    

lemma check_poly_weak_mono: fixes p :: "('v,'a :: poly_carrier)poly" and type :: 'w
  assumes "check_poly_weak_mono p" shows  "poly_weak_mono type p"
unfolding poly_weak_mono_def
proof (intro allI impI)
  fix f g :: "'v \<Rightarrow> ('w,'a)poly"
  assume fg: "\<forall> x. f x \<ge>p g x \<and> g x \<ge>p zero_poly"
  show "poly_subst f p \<ge>p poly_subst g p"
    unfolding poly_ge_def
  proof (intro allI impI)
    fix \<alpha> :: "('w,'a)assign"
    assume pos: "pos_assign \<alpha>"
    show "eval_poly \<alpha> (poly_subst f p) \<ge> eval_poly \<alpha> (poly_subst g p)"
    proof (simp only: poly_subst)
      let ?f = "\<lambda> v. eval_poly \<alpha> (f v)"
      let ?g = "\<lambda> v. eval_poly \<alpha> (g v)"
      have fg: "\<And> v. f v \<ge>p g v" and g: "\<And> v. g v \<ge>p  zero_poly"  using fg  by auto
      from assms have "\<And> m c. (m,c) \<in> set p \<Longrightarrow> ge c 0" unfolding check_poly_weak_mono_def by (auto simp: list_all_iff)
      thus "eval_poly ?f p \<ge> eval_poly ?g p"
      proof (induct p, simp add: ge_refl)
        case (Cons mc p)
        show ?case 
        proof (cases mc)
          case (Pair m c)
          with Cons have c: "c \<ge> 0" by auto
          from monom_subst_mono[OF pos fg g]
          have "(eval_monom ?f m \<ge> eval_monom ?g m)"  by force
          from times_left_mono[OF c this] have one: "eval_monom ?f m * c \<ge> eval_monom ?g m * c" .
          from Cons have two: "eval_poly ?f p \<ge> eval_poly ?g p" by auto
          from plus_left_mono[OF two] have "\<And> z. z + eval_poly ?f p \<ge> z + eval_poly ?g p" by (auto simp: field_simps)
          from ge_trans[OF plus_left_mono[OF one] this]
          show ?thesis by (simp add: Pair) 
        qed
      qed
    qed
  qed
qed

lemma check_poly_weak_mono_pos: 
  assumes "check_poly_weak_mono p" shows  "p \<ge>p zero_poly"
unfolding zero_poly_def
proof (rule check_poly_ge)
  from assms have "\<And> m c. (m,c) \<in> set p \<Longrightarrow> ge c 0" unfolding check_poly_weak_mono_def by (auto simp: list_all_iff)
  thus "check_poly_ge p []"
    by (induct p, simp add: check_poly_ge.simps,  clarify, auto simp: check_poly_ge.simps)
qed


(* extracts the coefficient for a given monomial and return the remaining polynomial *)
definition poly_split :: "('v monom) \<Rightarrow> ('v,'a :: zero)poly \<Rightarrow> 'a \<times> ('v,'a)poly" 
  where "poly_split m p \<equiv> case extract (\<lambda> (n,_). m =m n) p of None \<Rightarrow> (0,p) | Some (p1,(_,c),p2) \<Rightarrow> (c, p1 @ p2)"

lemma poly_split: assumes "poly_split m p = (c,q)"
  shows "p =p (m,c) # q"
proof (cases "extract (\<lambda> (n,_). m =m n) p")
  case None
  with assms have "(c,q) = (0,p)" unfolding poly_split_def by auto
  thus ?thesis unfolding eq_poly_def by auto
next
  case (Some res)
  obtain p1 mc p2 where "res = (p1,mc,p2)" by (cases res, auto)
  with extract_Some[OF Some[simplified this]] obtain a m' where p: "p = p1 @ (m',a) # p2" and m': "m =m m'" and res: "res = (p1,(m',a),p2)" by (cases mc, auto)
  from Some res assms have c: "c = a" and q: "q = p1 @ p2" unfolding poly_split_def by auto
  show ?thesis unfolding eq_poly_def by (simp add: p c q eq_monom[OF m'] field_simps)
qed 

lemma poly_split_eval: assumes "poly_split m p = (c,q)" 
  shows "eval_poly \<alpha> p = (eval_monom \<alpha> m * c) + eval_poly \<alpha> q"
using poly_split[OF assms] unfolding eq_poly_def by auto

definition check_poly_gt :: "('a \<Rightarrow> 'a \<Rightarrow> bool) \<Rightarrow> ('v,'a :: ordered_semiring_0)poly \<Rightarrow> ('v,'a)poly \<Rightarrow> bool"
where "check_poly_gt gt p q \<equiv> let (a1,p1) = poly_split [] p; (b1,q1) = poly_split [] q in gt a1 b1 \<and> check_poly_ge p1 q1"

definition check_monom_strict_mono :: "bool \<Rightarrow> 'v monom \<Rightarrow> 'v \<Rightarrow> bool"
  where "check_monom_strict_mono pm m v \<equiv> m \<noteq> [] \<and> tl m = [] \<and> fst (hd m) = v \<and> (\<lambda> p. if pm then p \<ge> 1 else p = 1) (snd (hd m))"

definition check_poly_strict_mono :: "bool \<Rightarrow> ('v,'a :: poly_carrier)poly \<Rightarrow> 'v \<Rightarrow> bool"
  where "check_poly_strict_mono pm p v \<equiv> list_ex (\<lambda> (m,c). (c \<ge> 1) \<and> check_monom_strict_mono pm m v) p"



locale poly_order_carrier = SN_one_mono_ordered_semiring_1 default gt 
  for default :: "'a :: poly_carrier" and gt +
  fixes power_mono :: "bool"
  assumes times_gt_mono: "\<lbrakk>gt y z; x \<ge> 1\<rbrakk> \<Longrightarrow> gt (y * x) (z * x)"
  and gt_imp_ge: "gt x y \<Longrightarrow> ge x y"
  and power_mono: "power_mono \<Longrightarrow> gt x y \<Longrightarrow> ge y 0 \<Longrightarrow> n \<ge> 1 \<Longrightarrow> gt (x ^ n) (y ^ n)"
begin


definition poly_gt :: "('v,'a)poly \<Rightarrow> ('v,'a)poly \<Rightarrow> bool" (infix ">p" 51)
where "p >p q = (\<forall> \<alpha>. pos_assign \<alpha> \<longrightarrow> gt (eval_poly \<alpha> p) (eval_poly \<alpha> q))"

lemma poly_gt_imp_poly_ge: "p >p q \<Longrightarrow> p \<ge>p q" unfolding poly_ge_def poly_gt_def using gt_imp_ge by blast

definition poly_strict_mono :: "'w \<Rightarrow> ('v,'a)poly \<Rightarrow> 'v \<Rightarrow> bool"
  where "poly_strict_mono type p v \<equiv> \<forall> f g :: ('v \<Rightarrow> ('w,'a)poly). (\<forall> w. (v \<noteq> w \<longrightarrow> f w = g w) \<and> g w \<ge>p zero_poly) \<and> (f v >p g v) \<longrightarrow> poly_subst f p >p poly_subst g p"

lemma check_monom_strict_mono: fixes \<alpha> \<beta> :: "('v,'a)assign" and v :: 'v and m :: "'v monom"
  assumes check: "check_monom_strict_mono power_mono m v"
  and gt: "gt (\<alpha> v) (\<beta> v)"
  and ge: "ge (\<beta> v) 0"
  shows "gt (eval_monom \<alpha> m) (eval_monom \<beta> m)"
proof (cases power_mono)
  case False
  with check have "m = [(v,1)]" unfolding check_monom_strict_mono_def by (cases m, cases "tl m", cases "hd m", auto)
  with gt show ?thesis by (auto)
next
  case True
  with check obtain n where m: "m = [(v,n)]" and n: "1 \<le> n" unfolding check_monom_strict_mono_def by (cases m, cases "tl m", cases "hd m", auto)
  from power_mono[OF True gt ge n] m show ?thesis by auto
qed

lemma check_poly_strict_mono: fixes type :: 'w
  assumes check1: "check_poly_strict_mono power_mono p v"
  and check2: "check_poly_weak_mono p"
  shows "poly_strict_mono type p v"
unfolding poly_strict_mono_def
proof (intro allI impI, clarify)
  fix f g :: "'b \<Rightarrow> ('w,'a)poly"
  assume fgw: "\<forall> w. (v \<noteq> w \<longrightarrow> f w = g w) \<and> g w \<ge>p zero_poly" and fgv: "f v >p g v"
  from fgw poly_gt_imp_poly_ge[OF fgv] have fgw2: "\<forall> w. f w \<ge>p g w \<and> g w \<ge>p zero_poly" using poly_ge_refl by force
  show "poly_subst f p >p poly_subst g p"
    unfolding poly_gt_def
  proof (intro allI impI)
    fix \<alpha> :: "('w,'a)assign"
    assume pos: "pos_assign \<alpha>"
    let ?e = "\<lambda> f p. eval_poly \<alpha> (poly_subst f p)"
    show "gt (?e f p) (?e g p)"
      using check1[unfolded check_poly_strict_mono_def, simplified list_ex_iff]
        check2[unfolded check_poly_weak_mono_def, simplified list_all_iff, THEN bspec]
    proof (induct p, simp)
      case (Cons mc p)
      obtain m c where mc: "mc = (m,c)" by (cases mc, auto)
      show ?case 
      proof (cases "(c \<ge> 1) \<and> check_monom_strict_mono power_mono m v")
        case True
        hence c: "c \<ge> 1" and m: "check_monom_strict_mono power_mono m v" by blast+
        have "gt (eval_poly \<alpha> (monom_subst f m)) (eval_poly \<alpha> (monom_subst g m))" (is "gt ?emf ?emg")
          by (simp add: monom_subst.simps one_poly_def, rule check_monom_strict_mono[OF m], 
              rule fgv[unfolded poly_gt_def, THEN spec, THEN mp[OF _ pos]], 
              rule fgw[unfolded poly_ge_def zero_poly_def, THEN spec, of v, simplified, THEN spec, THEN mp[OF _ pos]])
        from times_gt_mono[OF this c] have gt: "gt (c * ?emf) (c * ?emg)" by (auto simp: field_simps)
        from Cons(3) have "check_poly_weak_mono p" unfolding check_poly_weak_mono_def list_all_iff by auto
        from check_poly_weak_mono[OF this, unfolded poly_weak_mono_def, THEN spec, THEN spec, THEN mp[OF _ fgw2]]
        have ge: "poly_subst f p \<ge>p poly_subst g p" by auto
        show ?thesis by (simp add: mc poly_subst.simps del: monom_subst, rule compat2[OF plus_gt_left_mono[OF gt]],
          simp only: add_commute[of "c * ?emg" _], rule plus_left_mono, rule ge[unfolded poly_ge_def, THEN spec, THEN mp[OF _ pos]])
      next
        case False
        with Cons(2) mc have "\<exists> mc \<in> set p. (\<lambda> (m,c). (c \<ge> 1) \<and> check_monom_strict_mono power_mono m v) mc" by auto
        from Cons(1)[OF this] Cons(3) have rec: "gt (?e f p) (?e g p)" by simp
        from Cons(3) mc have c: "c \<ge> 0" by auto
        have "ge (eval_poly \<alpha> (monom_subst f m)) (eval_poly \<alpha> (monom_subst g m))" (is "ge ?emf ?emg")
          using monom_subst_mono[OF pos, of f g] fgw2 by blast
        from times_left_mono[OF c this] have ge: "c * ?emf \<ge> c * ?emg" by (auto simp: field_simps)
        show ?thesis by (simp add: mc poly_subst.simps del: monom_subst, simp only: field_simps, rule
           compat2[OF plus_gt_left_mono[OF rec] plus_left_mono[OF ge, simplified add_commute[of "c * ?emf"] add_commute[of "c * ?emg"]]])
      qed
    qed
  qed
qed
        
      

lemma check_poly_gt: 
  fixes p :: "('v,'a)poly"
  assumes "check_poly_gt gt p q" shows "p >p q"
proof -
  obtain a1 p1 where p: "poly_split [] p = (a1,p1)" by (cases "poly_split [] p", auto)
  obtain b1 q1 where q: "poly_split [] q = (b1,q1)" by (cases "poly_split [] q", auto)
  from p q assms have gt: "gt a1 b1" and ge: "p1 \<ge>p q1" unfolding check_poly_gt_def using check_poly_ge[of p1 q1]  by auto
  show ?thesis
  proof (unfold poly_gt_def, intro impI allI)
    fix \<alpha> :: "('v,'a)assign"
    assume "pos_assign \<alpha>"
    with ge have ge: "ge (eval_poly \<alpha> p1) (eval_poly \<alpha> q1)" unfolding poly_ge_def by simp
    from plus_gt_left_mono[OF gt] compat[OF plus_left_mono[OF ge]] have gt: "gt (a1 + eval_poly \<alpha> p1) (b1 + eval_poly \<alpha> q1)" by (force simp: field_simps)
    show "gt (eval_poly \<alpha> p) (eval_poly \<alpha> q)"
      by (simp add: poly_split[OF p, unfolded eq_poly_def] poly_split[OF q, unfolded eq_poly_def] gt)
  qed
qed

abbreviation poly_GT :: "('v,'a)poly ars"
where "poly_GT \<equiv> {(p,q) | p q. p >p q \<and> q \<ge>p zero_poly}"

lemma poly_compat: "\<lbrakk>p1 \<ge>p p2; p2 >p p3\<rbrakk> \<Longrightarrow> p1 >p p3"
using assms unfolding poly_ge_def poly_gt_def using compat by blast

lemma poly_compat2: "\<lbrakk>p1 >p p2; p2 \<ge>p p3\<rbrakk> \<Longrightarrow> p1 >p p3"
using assms unfolding poly_ge_def poly_gt_def using compat2 by blast

lemma poly_GT_SN: "SN poly_GT"
proof
  fix f
  assume f: "\<forall> i. (f i, f (Suc i)) \<in> poly_GT"
  have pos: "pos_assign ((\<lambda> x. 0) :: ('v,'a)assign)" (is "pos_assign ?ass") unfolding pos_assign_def using ge_refl by auto
  obtain g where g: "\<And> i. g i = eval_poly ?ass (f i)" by auto
  from f pos have "\<forall> i. ge (g (Suc i)) 0 \<and> gt (g i) (g (Suc i))" unfolding poly_gt_def g using pos_assign_poly by auto
  with SN show False unfolding SN_defs by blast 
qed

lemma poly_add_gt_mono: assumes "p1 >p p2" shows "poly_add p1 q >p poly_add p2 q"
using assms unfolding poly_gt_def by (auto simp: field_simps plus_gt_left_mono)

lemma poly_mult_gt_mono: 
  fixes q :: "('v,'a)poly"
  assumes gt: "p1 >p p2" and mono: "q \<ge>p one_poly"
  shows "poly_mult p1 q >p poly_mult p2 q"
proof (unfold poly_gt_def, intro impI allI, simp)
  fix \<alpha> :: "('v,'a)assign"
  assume p: "pos_assign \<alpha>"
  with gt have gt: "gt (eval_poly \<alpha> p1) (eval_poly \<alpha> p2)" unfolding poly_gt_def by simp
  from mono p have one: "eval_poly \<alpha> q \<ge> 1" unfolding poly_ge_def one_poly_def by auto
  show "gt (eval_poly \<alpha> p1 * eval_poly \<alpha> q) (eval_poly \<alpha> p2 * eval_poly \<alpha> q)"
    using times_gt_mono[OF gt one] .
qed
end


end